import inspect
import re
import sys
import os
import subprocess
from traceback import format_exc as py_format_exception
from json import dumps as json_encode
from json import loads as json_decode
from copy import deepcopy
from time import sleep
from typing import List, Callable, Union
from getpass import getpass
from contextlib import contextmanager
from datetime import datetime
from ..exception import InterruptExecution


this = sys.modules[__name__]
this.IS_CAPTURING_DESCRIPTORS = False

LEVEL_INTERNAL = 999
LEVEL_DEBUG = 37
LEVEL_INFO = 36
LEVEL_WARNING = 33
LEVEL_ERROR = 31
LEVEL_FATAL = 20

LOG_LEVELS = {
    'internal': LEVEL_INTERNAL,
    'debug': LEVEL_DEBUG,
    'info': LEVEL_INFO,
    'warning': LEVEL_WARNING,
    'error': LEVEL_ERROR,
    'fatal': LEVEL_FATAL
}

LOG_LEVEL_FORMATTING_MAPPING = {
    'internal': "\x1B[0m%TEXT%\x1B[0m",
    'debug':    "\x1B[0m%TEXT%\x1B[0m",
    'info':     "\x1B[1m%TEXT%\x1B[0m",
    'warn':     "\x1B[93m%TEXT%\x1B[0m",
    'error':    "\x1B[91m%TEXT%\x1B[0m",
    'fatal':    "\x1B[91m\x1B[5m%TEXT%\x1B[0m"
}

OUTPUT_PROCESSOR_CALLABLE_DEF = Callable[[Union[str, bytes], str], Union[str, bytes]]


class StandardOutputReplication(object):
    _out_streams: list
    _fileno: int

    def __init__(self, out_streams: list, fileno: int = None):
        self._out_streams = out_streams
        self._fileno = fileno

    def write(self, buf):
        for stream in self._out_streams:
            try:
                stream.write(buf)
            except TypeError:
                try:
                    stream.write(buf.encode('utf-8'))
                except AttributeError:
                    stream.write(str(buf))

        self.flush()

    def fileno(self):
        return self._fileno

    def flush(self):
        pass


class IO:
    """ Interacting with input and output - stdout/stderr/stdin, logging """

    silent = False
    log_level = LEVEL_INFO
    output_processors: List[OUTPUT_PROCESSOR_CALLABLE_DEF]

    def __init__(self):
        self.output_processors = []

    @contextmanager
    def capture_descriptors(self, target_files: List[str] = None, stream=None, enable_standard_out: bool = True):
        """Capture stdout and stderr from a block of code - use with 'with'"""

        if target_files is None:
            target_files = []

        if this.IS_CAPTURING_DESCRIPTORS:
            self.debug('Deep call to capture_descriptors()')

        this.IS_CAPTURING_DESCRIPTORS = True

        sys_stdout = sys.stdout
        sys_stderr = sys.stderr
        log_files = []

        outputs_stdout = []
        outputs_stderr = []

        # 1. Prepare standard out/err
        if enable_standard_out:
            outputs_stdout.append(sys_stdout)
            outputs_stderr.append(sys_stderr)

        # 2. Prepare logs
        for target_file in target_files:
            subprocess.call(['mkdir', '-p', os.path.dirname(target_file)])

            log_file = open(target_file, 'wb', buffering=0)
            log_file.no_flush = True
            log_files.append(log_file)

            outputs_stdout.append(log_file)
            outputs_stderr.append(log_file)

        # 3. Prepare StringIO
        if stream:
            outputs_stdout.append(stream)
            outputs_stderr.append(stream)

        # 4. Mock
        sys.stdout = StandardOutputReplication(outputs_stdout, sys.stdout.fileno())
        sys.stderr = StandardOutputReplication(outputs_stderr, sys.stderr.fileno())

        # 5. Action!
        yield

        # 6. Revert standard out/err
        sys.stdout = sys_stdout
        sys.stderr = sys_stderr

        # 7. Clean up: close all log files
        for log_file in log_files:
            log_file.close()

        this.IS_CAPTURING_DESCRIPTORS = False

    def inherit_silent(self, io: 'SystemIO'):
        self.silent = io.is_silent(consider_ui=False)

    def is_silent(self) -> bool:
        """Is output silent? In silent mode OPTIONAL MESSAGES are not shown"""

        return self.silent

    #
    # Log level - mutable setting
    #
    def set_log_level(self, desired_level_name: str):
        if desired_level_name not in LOG_LEVELS:
            raise Exception('Invalid log level name')

        self.log_level = LOG_LEVELS[desired_level_name]

    def is_log_level_at_least(self, log_level: str) -> bool:
        return self.log_level >= LOG_LEVELS[log_level]

    def get_log_level(self) -> str:
        for name, severity in LOG_LEVELS.items():
            if severity == self.log_level:
                return name

        raise Exception('Log level not set')

    #
    # Standard output/error
    #

    def _stdout(self, text):
        sys.stdout.write(text)

    def _stderr(self, text):
        sys.stderr.write(text)

    def out(self, text):
        """ Standard output """
        self._stdout(self._process_output(text, 'stdout'))

    def outln(self, text):
        """ Standard output + newline """
        self.out(text + "\n")

    def err(self, text):
        """ Standard error """
        self._stderr(self._process_output(text, 'stderr'))

    def errln(self, text):
        """ Standard error + newline """
        self.err(text + "\n")

    def opt_errln(self, text):
        """ Optional errln() """

        if not self.is_silent():
            self.errln(text)

    def opt_out(self, text):
        """ Optional output - fancy output skipped in --silent mode """

        if not self.is_silent():
            self.out(text)

    def opt_outln(self, text):
        """ Optional output - fancy output skipped in --silent mode + newline """

        if not self.is_silent():
            self.outln(text)

    #
    # Logs
    #

    def internal(self, text):
        """Logger: internal
           Should be used only by RKD core for more intensive logging
        """
        if self.log_level < LEVEL_INTERNAL:
            return

        text = inspect.stack()[1][3] + ' ~> ' + text
        self.log(text, 'internal')

    def debug(self, text):
        """Logger: debug

        """
        if self.log_level >= LEVEL_DEBUG:
            self.log(text, 'debug')

    def info(self, text):
        """Logger: info

        """

        if self.log_level >= LEVEL_INFO:
            self.log(text, 'info')

    def warn(self, text):
        """Logger: warn

        """

        if self.log_level >= LEVEL_WARNING:
            self.log(text, 'warn')

    def error(self, text):
        """Logger: error

        """

        if self.log_level >= LEVEL_ERROR:
            self.err_log(text, 'error')

    def critical(self, text):
        """Logger: critical

        """

        if self.log_level >= LEVEL_FATAL:
            self.err_log(text, 'critical')

    def log(self, text, level: str):
        if not self.is_silent():
            self.outln(self._format_log(text, level))

    def err_log(self, text, level: str):
        if not self.is_silent():
            self.errln(self._format_log(text, level))

    def _format_log(self, text, level: str) -> str:
        current_time = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        level = LOG_LEVEL_FORMATTING_MAPPING[level].replace('%TEXT%', level)

        return "\x1B[2m[%s]\x1B[0m[%s]: \x1B[2m%s\x1B[0m" % (current_time, level, text)

    def print_group(self, text):
        """Prints a colored text inside brackets [text] (optional output)

        """

        self.opt_outln("\x1B[33m[%s]\x1B[0m" % text)

    #
    # Lines and separators
    #

    def print_line(self):
        """Prints a newline

        """

        self.outln('')

    def print_opt_line(self):
        """Prints a newline (optional output)
        """

        self.opt_outln('')

    def print_separator(self):
        """Prints a text separator (optional output)
        """
        self.opt_outln("\x1B[37m%s\x1B[0m" % '-----------------------------------')

    #
    #  Statuses
    #

    def success_msg(self, text):
        """Success message (optional output)
        """

        self.opt_outln("\x1B[92m%s\x1B[0m" % text)

    def error_msg(self, text):
        """Error message
        """

        self.errln("\x1B[91m%s\x1B[0m" % text)

    def warn_msg(self, text) -> None:
        """Warning message (optional output)"""

        self.opt_outln("\x1B[33m%s\x1B[0m" % text)

    def info_msg(self, text):
        """Informational message (optional output)
        """

        self.opt_outln("\x1B[93m%s\x1B[0m" % text)

    #
    # Standard formatting
    #
    def h1(self, text):
        """Heading #1 (optional output)
        """

        self.opt_outln("\x1B[93m  ##> %s\x1B[0m" % text)

    def h2(self, text):
        """Heading #2 (optional output)
        """

        self.opt_outln("\x1B[93m   ===> %s\x1B[0m" % text)

    def h3(self, text):
        """Heading #3 (optional output)
        """

        self.opt_outln("\x1B[33m    --> %s\x1B[0m" % text)

    def h4(self, text):
        """Heading #3 (optional output)
        """

        self.opt_outln("\x1B[33m     ... %s\x1B[0m" % text)

    def add_output_processor(self, callback: OUTPUT_PROCESSOR_CALLABLE_DEF):
        """
        Registers a output processing callback
        Each byte outputted by this IO instance will go through a set of registered processors

        Example use cases:
            - Hide sensitive information (secrets)
            - Reformat output
            - Strip long stdouts from commands
            - Change colors
            - Add/remove formatting

        :param callback:
        :return:
        """

        self.output_processors.append(callback)

    def _process_output(self, text, origin: str):
        """
        Process output by passing it through multiple registered processors
        :param text:
        :param origin:
        :return:
        """

        for txt_filter in self.output_processors:
            try:
                processed = txt_filter(text, origin)

                if type(processed) != str:
                    raise Exception('Output processor must return a str')

                text = processed

            # do not allow exceptions in core output buffering module, unless we are debugging
            except Exception:
                if self.log_level >= LEVEL_DEBUG:
                    raise

                pass

        return text


class SystemIO(IO):
    """ Used for logging outside of tasks """

    _ui = True

    def capture_descriptors(self, target_file: str = None, stream=None, enable_standard_out: bool = True):
        pass

    def set_display_ui(self, ui: bool):
        self._ui = ui

    def is_silent(self, consider_ui: bool = True) -> bool:
        if consider_ui and not self._ui:
            return True

        return self.silent


class NullSystemIO(SystemIO):
    def _stdout(self, text):
        pass

    def _stderr(self, text):
        pass


class BufferedSystemIO(SystemIO):
    _buffer = ''

    def _stdout(self, text):
        self._buffer += text

    def _stderr(self, text):
        self._buffer += text

    def get_value(self):
        return self._buffer

    def clear_buffer(self):
        self._buffer = ''


class Wizard(object):
    _max_retries: int = 3
    answers: dict
    io: 'IO'
    task: 'TaskInterface'
    to_env: dict
    sleep_time = 1
    filename: str

    def __init__(self, task: 'TaskInterface', filename: str = 'tmp-wizard.json'):
        self.answers = {}
        self.task = task
        self.io = task.io()
        self.to_env = {}
        self.filename = filename

    def ask(self, title: str, attribute: str, regexp: str = '', to_env: bool = False, default: str = None,
            choices: list = [], secret: bool = False) -> 'Wizard':
        """Asks user a question

        Usage:
            .. code:: python

                wizard = Wizard(self)
                wizard.ask('In which year the Spanish social revolution has begun?',
                           attribute='year',
                           choices=['1936', '1910'])
                wizard.finish()
        """

        retried = 0
        value = None
        full_text_to_ask = title

        if choices and regexp:
            raise Exception('Please choose between regexp and choices validation.')

        if regexp:
            full_text_to_ask += " [%s]" % regexp

        if default:
            full_text_to_ask += " [default: %s]" % default

        if choices:
            full_text_to_ask += " [%s]" % ', '.join(choices)

        full_text_to_ask += ": "

        while value is None or not self.is_valid(value, regexp, choices):
            self.io.out(full_text_to_ask + "\n -> ")
            value = self.input(secret=secret)

            if default and not value.strip():
                value = default

            if retried >= self._max_retries:
                raise InterruptExecution('Invalid value given')

            if self.is_valid(value, regexp):
                break

            retried += 1
            sleep(self.sleep_time)

        if to_env:
            self.to_env[attribute] = value
            return self

        self.answers[attribute] = value
        return self

    def input(self, secret: bool = False):
        """Extracted for unit testing to be possible easier"""

        if os.getenv('__WIZARD_INPUT'):
            return os.getenv('__WIZARD_INPUT')

        if secret:
            return getpass(prompt='')

        return input()

    @staticmethod
    def is_valid(value: any, regexp: str = '', choices: list = []):
        if choices:
            return value in choices

        if regexp:
            return re.match(regexp, value) is not None

        return True

    def load_previously_stored_values(self):
        """Load previously saved values"""

        if os.path.isfile('.rkd/' + self.filename):
            with open('.rkd/' + self.filename, 'rb') as f:
                self.answers = json_decode(f.read())

        self.to_env = deepcopy(os.environ)

    def finish(self) -> 'Wizard':
        """Commit all pending changes into json and .env files"""

        self.io.info('Writing to .rkd/' + self.filename)
        with open('.rkd/' + self.filename, 'wb') as f:
            f.write(json_encode(self.answers).encode('utf-8'))

        self.io.info('Writing to .env')
        for attribute, value in self.to_env.items():
            self.task.rkd(
                [':env:set', '--name="%s"' % attribute, '--value="%s"' % value],
                verbose=False,
                capture=True
            )

        return self


def clear_formatting(text: str) -> str:
    text = re.sub("\\x1B\[([0-9]+)m", '', text)

    return text


def output_formatted_exception(exc: Exception, title: str, io: IO):
    """Formats a catched exception and displays as user-friendly by default (without a stack trace)

    When at least there is a "debug" level of error reporting, then an original stack trace would be displayed
    Everything goes through the RKD's IO, not directly to the stdout/stderr. The stderr is used there naturally.
    """

    try:
        io.errln('During "%s" a critical error happened' % title)
        io.print_line()

        if io.is_log_level_at_least('debug'):
            io.errln(py_format_exception())
            return

        cause = exc.__cause__

        while cause is not None:
            io.errln('\x1B[91m(Caused by) %s: \x1B[93m%s\x1B[0m' % (
                cause.__class__.__name__,
                indent_new_lines(str(cause), 4)
            ))

            cause = cause.__cause__

        io.errln('\x1B[91m%s: \x1B[93m%s\x1B[0m' % (exc.__class__.__name__, indent_new_lines(str(exc), 4)))
        io.print_line()
        io.errln('\x1B[37mRetry with "-rl debug" switch before failed task to see stacktrace\x1B[0m')

    except Exception:
        print('FATAL: During exception formatting there was an unrecoverable error')
        print(py_format_exception())
        sys.exit(1)

    return


def indent_new_lines(text: str, num: int = 4):
    """Inserts spaces at the beginning of each new line"""

    return text.replace("\n", "\n" + (" " * num))


class UnbufferedStdout(object):
    """Executes flush() after each write"""

    def __init__(self, stream):
        self.stream = stream

    def write(self, data):
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas):
        self.stream.writelines(datas)
        self.stream.flush()

    def fileno(self):
        return self.stream.fileno()

    def __getattr__(self, attr):
        return getattr(self.stream, attr)


def get_environment_copy() -> dict:
    """Copy environment variables keeping the values escaped

    Dollar character escaping assumptions:
        If the dollar char is present in variable, then it means that it was escaped before
        if it would not be escaped before passing to RKD then it would be EVALUATED.

        So we keep it escaped.
    """

    return dict(
        map(
            lambda kv: (kv[0], str(kv[1]).replace('$', '\$')),
            dict(deepcopy(os.environ)).items()
        )
    )




from typing import List, Callable, Union, Optional
from .argparsing.model import TaskArguments, ArgumentBlock
from .api.syntax import TaskDeclaration, GroupDeclaration
from .context import ApplicationContext
from .exception import InterruptExecution, \
    ExecutionRetryException, \
    ExecutionErrorActionException, \
    TaskNotFoundException, ExecutionRescueException, ExecutionRescheduleException
from .aliasgroups import AliasGroup


CALLBACK_DEF = Callable[[TaskDeclaration, int, Union[GroupDeclaration, None], list], None]


class TaskResolver(object):
    """
    Responsible for finding all tasks and:
        - expanding groups (flatten tasks)
        - connecting each task to parent
        - preserve valid order of task validation/execution
    """

    _ctx: ApplicationContext
    _alias_groups: List[AliasGroup]

    def __init__(self, ctx: ApplicationContext, alias_groups: List[AliasGroup]):
        self._ctx = ctx
        self._alias_groups = alias_groups

    def resolve(self, requested_blocks: List[ArgumentBlock], callback: CALLBACK_DEF):
        """
        Iterate over flatten list of tasks, one by one and call a callback for each task

        :param requested_blocks:
        :param callback:
        :return:
        """

        task_num = 0

        for block in requested_blocks:
            for task_request in block.tasks():
                task_num += 1

                try:
                    self._resolve_element(task_request, callback, task_num, block)
                except InterruptExecution:
                    return

    def _resolve_name_from_alias(self, task_name: str) -> Optional[str]:
        """Resolves task group's shortcuts eg. :hb -> :harbor"""

        for alias in self._alias_groups:
            resolved = alias.append_alias_to_task(task_name)

            if resolved:
                return resolved

    def _resolve_elements(self, requests: List[TaskArguments], callback: CALLBACK_DEF, task_num: int,
                          block: ArgumentBlock) -> None:

        for request in requests:
            self._resolve_element(request, callback, task_num, block)

    def _resolve_element(self, task_request: TaskArguments, callback: CALLBACK_DEF, task_num: int,
                         block: ArgumentBlock) -> None:

        """Checks task by name if it was defined in context, if yes then unpacks declarations and prepares callbacks"""

        self._ctx.io.internal('Resolving {}'.format(task_request))

        try:
            # @todo: Possibly clone required - shell summary shows only 2 tasks executed, when there were executed more but of same type
            ctx_declaration = self._ctx.find_task_by_name(task_request.name())

        # maybe a task name is an alias to other task defined by alias groups
        except TaskNotFoundException:
            task_from_alias = self._resolve_name_from_alias(task_request.name())

            if not task_from_alias:
                raise

            ctx_declaration = self._ctx.find_task_by_name(task_from_alias)

        self._ctx.io.internal('Resolved as {}'.format(ctx_declaration))

        if isinstance(ctx_declaration, TaskDeclaration):
            declarations: List[TaskDeclaration] = ctx_declaration.to_list()
            parent = None
        elif isinstance(ctx_declaration, GroupDeclaration):
            declarations: List[TaskDeclaration] = ctx_declaration.get_declarations()
            parent = ctx_declaration
        else:
            raise Exception('Cannot resolve task - unknown type "%s"' % str(ctx_declaration))

        # connect TaskDeclaration to blocks for context
        declaration_num = 0
        for declaration in declarations:
            declaration: TaskDeclaration
            declarations[declaration_num] = declaration.with_connected_block(block)
            declaration_num += 1

        self._iterate_over_declarations(callback, declarations, task_num, parent, task_request)

    def _iterate_over_declarations(self, callback: CALLBACK_DEF, declarations: list, task_num: int,
                                   parent: Optional[GroupDeclaration], task_request: TaskArguments):

        """Recursively go through all tasks in correct order, executing a callable on each"""

        for declaration in declarations:
            declaration: TaskDeclaration

            if isinstance(declaration, GroupDeclaration):
                self._iterate_over_declarations(
                    callback, declaration.get_declarations(),
                    task_num, declaration, task_request
                )
                continue

            try:
                callback(
                    declaration,
                    task_num,
                    parent,
                    # the arguments there will be mixed in order:
                    #  - first: defined in Makefile
                    #  - second: commandline arguments
                    #
                    #  The argparse in Python will take the second one as priority.
                    #  We do not try to remove duplications there to not increase complexity
                    #  of the solution - it works now.
                    declaration.get_args() + task_request.args()
                )

            #
            # Resolver is able to resolve dynamically additional fallback tasks on-demand
            # The status of the overall pipeline depends on decision of ProgressObserver, the resolver is only
            # resolving and scheduling tasks, not deciding about results
            #

            except ExecutionRetryException as exc:
                # multiple tasks to resolve, then retry
                if exc.args:
                    self._resolve_elements(
                        requests=exc.args,
                        callback=callback,
                        task_num=task_num,
                        block=ArgumentBlock.from_empty()
                    )
                    return

                # single task to retry
                self._iterate_over_declarations(
                    callback=callback,
                    declarations=[declaration],
                    task_num=task_num,
                    parent=parent,
                    task_request=task_request
                )

            except ExecutionRescheduleException as reschedule_action:
                self._resolve_elements(
                    requests=reschedule_action.tasks_to_schedule,
                    callback=callback,
                    task_num=task_num,
                    block=ArgumentBlock.from_empty()
                )



"""
Commandline Blocks
==================

Blocks are adding error handling to pipeline of tasks, allowing to define fallback actions.
Can be used in shell, in TaskAliases.

Example:

.. code:: shell

    :prepare '{@rescue :rollback @error :notify "Failed" @retry 3}' :deploy --env=test{/@}


Rules:

    - @error and @rescue cannot exist together
    - Blocks cannot be nested

Available modifiers:

    - @error (execute a task, when any of task fails, break the pipeline. Good for notifications)
    - @rescue (execute a task, when any of task fails, don't break the pipeline if rescue action succeeds)
    - @retry (retry failed task up to X times, good for unstable tasks as a workaround until those tasks are not fixed)
    - @retry-block (retry whole block in case, when a single task fails)
"""

import re
from typing import Tuple, List, Union, Dict
from .model import ArgumentBlock
from ..exception import CommandlineParsingError

TOKEN_SEPARATOR = ' '
TOKEN_BEGIN_BLOCK = '{@'
TOKEN_BEGIN_BLOCK_ENDING = '{/@'
TOKEN_CLOSING_BLOCK = '}'
TOKEN_BLOCK_REFERENCE_OPENING = '[[[$RKT_BLOCK'
TOKEN_BLOCK_REFERENCE_CLOSING = ']]]'
TEMPORARY_SEPARATOR = '[[[$_RKD_SEP]]]'
ALLOWED_MODIFIERS = ['rescue', 'error', 'retry', 'retry-block']


def strip_empty_elements(to_strip: list) -> list:
    try:
        if not to_strip[0]:
            del to_strip[0]
    except KeyError:
        pass

    try:
        if not to_strip[-1]:
            del to_strip[-1]
    except KeyError:
        pass

    return to_strip


def parse_blocks(commandline: List[str]) -> Tuple[List[str], dict]:
    """
    Parses commandline into blocks. Uses a cursor-based methodology

    Examples:
        :aaa {@rescue :rollback --env=test-2} :deploy --env=test {/@}
        :bbb {@rescue :rollback @error :notify "Failed" @retry 3}:deploy --env=test{/@}

        Given we have ":bbb {@rescue :rollback @error :notify "Failed" @retry 3}:deploy --env=test{/@}"
        Then we extract it into ":bbb [[[$_RKD_GROUP_1]]]" + list of objects [ArgumentBlock] with one element

    :author: dkwebbie <github.com/dkwebbie>
    """

    # TEMPORARY_SEPARATOR allows to keep original data structure, as manual, primitive parsing can lose quoted strings
    # for example
    commandline_as_str = TEMPORARY_SEPARATOR.join(commandline)

    cursor = 0
    cursor_end = len(commandline_as_str)
    block_id = 0
    collected_blocks = {}

    while cursor < cursor_end:
        opening_match = commandline_as_str.find(TOKEN_BEGIN_BLOCK, cursor)

        # 1. Find opening "{@"
        if opening_match >= 0:
            block_id += 1
            cursor = opening_match  # after "{@"

            closing_match = commandline_as_str.find(TOKEN_CLOSING_BLOCK, cursor)

            # if accidentally closing tag was matched, then we do not have a match
            if commandline_as_str[closing_match - 3:closing_match] == "{/@":
                closing_match = 0

            # 2. Find closing "}" of the "{@"
            if closing_match < 0:
                raise CommandlineParsingError.from_block_closing_not_found(cursor)

            cursor = closing_match + 1  # after "{@block}"
            header = commandline_as_str[opening_match:closing_match]
            inner_arguments = parse_block_header(header)

            # 3. Find {/@...} - block ending
            block_ending_content = TOKEN_BEGIN_BLOCK_ENDING + TOKEN_CLOSING_BLOCK
            block_ending_match = commandline_as_str.find(block_ending_content, cursor)

            if block_ending_match < 0:
                raise CommandlineParsingError.from_block_ending_not_found(block_ending_content)

            body = strip_empty_elements(
                commandline_as_str[closing_match + 1:block_ending_match].split(TEMPORARY_SEPARATOR)
            )
            cursor = block_ending_match + len(block_ending_content)  # after "{/@}"

            if TOKEN_BEGIN_BLOCK in str(body):
                raise CommandlineParsingError.from_nested_blocks_not_allowed(TOKEN_BEGIN_BLOCK, header)

            block_token = ((TOKEN_BLOCK_REFERENCE_OPENING + '%i' + TOKEN_BLOCK_REFERENCE_CLOSING) % block_id)
            commandline_as_str = commandline_as_str[0:opening_match] + \
                                 block_token + \
                                 commandline_as_str[block_ending_match + len(block_ending_content):]

            # difference - we replace {@...}{/@} with eg. [[[$RKT_BLOCK1]]] - how many characters were removed
            after_replace_difference = ((block_ending_match + len(block_ending_content)) - opening_match) - len(block_token)
            cursor = cursor - after_replace_difference  # we are RIGHT after [[[$RKT_BLOCK1]]]

            try:
                collected_blocks[block_token.strip()] = ArgumentBlock(body=body, **inner_arguments)
            except TypeError as e:
                raise CommandlineParsingError.from_block_unknown_modifier(header, e)

        else:
            cursor += 1

    return commandline_as_str.split(TEMPORARY_SEPARATOR), collected_blocks


def parse_block_header(block_header: str) -> Dict[str, Union[str, int]]:
    """
    Parses a header ex. "{@retry 3"

    :author: dkwebbie <github.com/dkwebbie>
    """

    parsed = re.findall('@([a-z\-]+)([^@]*)', block_header)
    as_dict = {}

    if not parsed:
        raise CommandlineParsingError.from_block_header_parsing_exception(block_header)

    for result in parsed:
        result: List[str]
        name = result[0].strip()

        if name not in ALLOWED_MODIFIERS:
            raise CommandlineParsingError.from_block_unknown_modifier(block_header,
                                                                      Exception('Unknown modifier "%s"' % name))

        if name in as_dict:
            raise CommandlineParsingError.from_block_modifier_declared_twice(name, block_header)

        as_dict[name.replace('-', '_')] = result[1].strip()

    return as_dict



from typing import List
from jsonschema import ValidationError
from .argparsing.model import TaskArguments


class ContextException(Exception):
    pass


class TaskNotFoundException(ContextException):
    pass


class TaskExecutionException(Exception):
    pass


class InterruptExecution(TaskExecutionException):
    pass


class ExecutionRetryException(TaskExecutionException):
    """Internal signal to retry a task"""

    args: List[TaskArguments]

    def __init__(self, args: List[TaskArguments] = None):
        if args is None:
            args = []

        self.args = args


class ExecutionRescheduleException(TaskExecutionException):
    """Internal signal to put extra task into resolve/schedule queue of TaskResolver"""

    tasks_to_schedule: List[TaskArguments]

    def __init__(self, tasks_to_schedule: List[TaskArguments]):
        self.tasks_to_schedule = tasks_to_schedule


class ExecutionRescueException(ExecutionRescheduleException):
    """Internal signal to call a rescue set of tasks in case of given task fails"""


class ExecutionErrorActionException(ExecutionRescheduleException):
    """Internal signal to call an error notification in case when given task fails"""


class TaskException(ContextException):
    pass


class UndefinedEnvironmentVariableUsageError(TaskException):
    pass


class EnvironmentVariableNotUsed(TaskException):
    pass


class EnvironmentVariableNameNotAllowed(TaskException):
    def __init__(self, var_name: str):
        super().__init__('Environment variable with this name "' + var_name + '" cannot be declared, it probably a' +
                         ' commonly reserved name by operating systems')


class UserInputException(Exception):
    pass


class BlockDefinitionLogicError(Exception):
    @staticmethod
    def from_both_rescue_and_error_defined():
        return BlockDefinitionLogicError('Block "{0:s}" cannot define both @rescue and @error'.format(task.block().body))


class NotSupportedEnvVariableError(UserInputException):
    pass


class YamlParsingException(ContextException):
    """Logic or syntax errors in makefile.yaml"""


class YAMLFileValidationError(YamlParsingException):
    """Errors related to schema validation"""

    def __init__(self, err: ValidationError):
        super().__init__('YAML schema validation failed at path "%s" with error: %s' % (
            '.'.join(list(map(str, list(err.path)))),
            str(err.message)
        ))


class ParsingException(ContextException):
    """Errors related to parsing YAML/Python syntax"""

    @classmethod
    def from_import_error(cls, import_str: str, class_name: str, error: Exception) -> 'ParsingException':
        return cls(
            'Import "%s" is invalid - cannot import class "%s" - error: %s' % (
                import_str, class_name, str(error)
            )
        )

    @classmethod
    def from_class_not_found_in_module_error(cls, import_str: str, class_name: str,
                                             import_path: str) -> 'ParsingException':
        return cls(
            'Import "%s" is invalid. Class or method "%s" not found in module "%s"' % (
                import_str, class_name, import_path
            )
        )


class DeclarationException(ContextException):
    """Something wrong with the makefile.py/makefile.yaml """


class ContextFileNotFoundException(ContextException):
    """When makefile.py, makefile.yaml, makefile.yml not found (at least one needed)"""

    def __init__(self, path: str):
        super().__init__('The directory "%s" should contain at least makefile.py, makefile.yaml or makefile.yml' % path)


class PythonContextFileNotFoundException(ContextFileNotFoundException):
    """When makefile.py is not found"""

    def __init__(self, path: str):
        super().__init__('Python context file not found at "%s"' % path)


class NotImportedClassException(ContextException):
    """When class was not imported"""

    def __init__(self, exc: ImportError):
        super().__init__(
            'Your Makefile contains a reference to not available or not installed Python module "%s"' % str(exc)
        )


class EnvironmentVariablesFileNotFound(ContextException):
    """.env file specified, but not existing"""

    def __init__(self, path: str, lookup_paths: list):
        super().__init__(
            'Specified file "%s" as environment variable provider does not exist. Looked in: %s' % (
                path, str(lookup_paths)
            )
        )


class RuntimeException(Exception):
    pass


class MissingInputException(RuntimeException):
    def __init__(self, arg_name: str, env_name: str):
        super().__init__('Either "%s" switch not used, either "%s" was not defined in environment' % (
            arg_name, env_name
        ))


class CommandlineParsingError(RuntimeException):
    @staticmethod
    def from_block_header_parsing_exception(block_header: str) -> 'CommandlineParsingError':
        return CommandlineParsingError('Cannot parse block header "{}"'.format(block_header))

    @staticmethod
    def from_block_modifier_declared_twice(name: str, block_header: str) -> 'CommandlineParsingError':
        return CommandlineParsingError('Cannot declare "{}" twice in block "{}'.format(name, block_header))

    @staticmethod
    def from_block_unknown_modifier(header: str, e: Exception) -> 'CommandlineParsingError':
        return CommandlineParsingError('Block "{}" contains invalid modifier, raised error: {}'.format(header, str(e)))

    @staticmethod
    def from_nested_blocks_not_allowed(token: str, header: str) -> 'CommandlineParsingError':
        return CommandlineParsingError('Nesting blocks "{}" not allowed, attempted inside block "{}"'
                                       .format(token, header))

    @staticmethod
    def from_block_closing_not_found(pos: int):
        return CommandlineParsingError('Parsing exception: Closing character "}" not found for {@ opened at %i' % pos)

    @staticmethod
    def from_block_ending_not_found(block: str):
        return CommandlineParsingError('Parsing exception: Block ending - %s not found' % block)



from typing import Union, Dict
from ..api.syntax import TaskDeclaration
from ..api.syntax import GroupDeclaration
from ..argparsing.model import ArgumentBlock
from ..inputoutput import SystemIO


STATUS_STARTED = 'started'
STATUS_ERRORED = 'errored'
STATUS_FAILURE = 'failure'
STATUS_SUCCEED = 'succeed'

"""
Can be treated as "succeed" because "in-rescue" means that we don't check task status, instead we start a new task
and check that new rescue-task status instead of failed task status
"""
STATUS_RESCUE_STATE = 'in-rescue'


class TaskResult(object):
    task: TaskDeclaration
    status: str

    def __init__(self, task: TaskDeclaration, status: str):
        self.task = task
        self.status = status

    def has_succeed(self) -> bool:
        return self.status in [STATUS_SUCCEED, STATUS_RESCUE_STATE]


class ProgressObserver(object):
    """
    Carefuly tracks tasks execution progress, have answers to questions such as:
        - were there any tasks failed?

    This service is a REGISTRY.
    """

    _io: SystemIO
    _executed_tasks: Dict[str, TaskResult]

    def __init__(self, io: SystemIO):
        self._io = io
        self._executed_tasks = {}

    @staticmethod
    def _format_parent_task(parent: Union[GroupDeclaration, None]) -> str:
        return ('[part of ' + parent.get_name() + ']') if parent else ''

    def task_started(self, declaration: TaskDeclaration, parent: Union[GroupDeclaration, None], args: list):
        """ When task is just started """

        self._executed_tasks[declaration.get_unique_id()] = TaskResult(declaration, STATUS_STARTED)

        self._io.info_msg(' >> Executing %s %s %s' % (
            declaration.to_full_name(),
            ' '.join(args),
            self._format_parent_task(parent)
        ))

    def task_errored(self, declaration: TaskDeclaration, exception: Exception):
        """ On exception catched in task execution """

        self._set_status(declaration, STATUS_ERRORED)

        self._io.print_opt_line()
        self._io.error_msg('The task "%s" was interrupted with an %s' % (
            declaration.to_full_name(),
            str(exception.__class__)
        ))
        self._io.print_separator()
        self._io.print_opt_line()

    def task_failed(self, declaration: TaskDeclaration, parent: Union[GroupDeclaration, None]):
        """ When task returns False """

        self._set_status(declaration, STATUS_FAILURE)

        if not declaration.get_task_to_execute().is_silent_in_observer():
            self._io.print_opt_line()
            self._io.error_msg('The task "%s" %s ended with a failure' % (
                declaration.to_full_name(),
                self._format_parent_task(parent)
            ))
            self._io.print_separator()
            self._io.print_opt_line()

    def task_succeed(self, declaration: TaskDeclaration, parent: Union[GroupDeclaration, None]):
        """ When task success """

        self._set_status(declaration, STATUS_SUCCEED)

        if not declaration.get_task_to_execute().is_silent_in_observer():
            self._io.print_opt_line()
            self._io.success_msg('The task "%s" %s succeed.' % (
                declaration.to_full_name(),
                self._format_parent_task(parent)
            ))
            self._io.print_separator()
            self._io.print_opt_line()

    def execution_finished(self):
        """
        When all tasks were executed - the TaskExecutor finished its job
        """

        if self.is_at_least_one_task_failing():
            self._io.error_msg('Execution failed with %i failed tasks of %i total tasks scheduled for execution' % (
                self.count_failed_tasks(), len(self._executed_tasks)
            ))
        else:
            self._io.success_msg('Successfully executed %i tasks.' % len(self._executed_tasks))

        self._io.print_opt_line()

    def _set_status(self, declaration: TaskDeclaration, status: str):
        """Internally mark given task as done + save status"""

        self._io.internal('{} task, unique_id={}, status={}'.format(str(declaration), declaration.get_unique_id(), status))
        self._executed_tasks[declaration.get_unique_id()] = TaskResult(declaration, status)

    def is_at_least_one_task_failing(self) -> bool:
        return self.count_failed_tasks() >= 1

    def count_failed_tasks(self) -> int:
        return len({
            k: v for k, v in self._executed_tasks.items() if not v.has_succeed()
        }.values())

    def group_of_tasks_retried(self, block: ArgumentBlock):
        """
        When a block failed and needs to be retried (even intermediate success steps)
        """

        executed_tasks_that_belongs_to_block = {
            k: v for k, v in self._executed_tasks.items() if v.task.block() is block
        }

        for declaration in executed_tasks_that_belongs_to_block.values():
            self.task_retried(declaration)

    def task_retried(self, declaration: TaskDeclaration):
        self._io.warn_msg('Task "{}" was retried'.format(declaration.to_full_name()))
        self._set_status(declaration, STATUS_STARTED)

    def task_rescue_attempt(self, declaration: TaskDeclaration):
        self._io.warn_msg('Task "{}" rescue attempt started'.format(declaration.to_full_name()))
        self._set_status(declaration, STATUS_RESCUE_STATE)



from copy import deepcopy
from typing import List, Dict


class TaskArguments(object):
    """
    Task name + commandline switches model
    """

    _name: str
    _args: list

    def __init__(self, task_name: str, args: list):
        self._name = task_name
        self._args = args

    def __repr__(self):
        return 'TaskCall<%s (%s)>' % (self._name, str(self._args))

    def name(self):
        return self._name

    def args(self):
        return self._args

    def with_args(self, new_args: list) -> 'TaskArguments':
        clone = deepcopy(self)
        clone._args = new_args

        return clone


class ArgumentBlock(object):
    """
    ArgumentBlock
    =============

    Stores information about construction of blocks:
        {@block @error :notify @retry 2}:task1 --param1=value1 :task2{/@block}

    Lifetime:
        - Initially could store *body* (raw string, from example: ":task1 --param1=value1 :task2")
        - Later parsers are filling up the *_tasks* attribute with parsed TaskArguments
        - At last stage the *RKD's Executor component* is reading from ArgumentBlock and deciding if task should be
          retried, if there should be any error handling. The *_retry_counter_per_task* and *_retry_counter_on_whole_block*
          fields are mutable to track the progress of error handling

    Notice: Fields like on_error, on_rescue are filled up after block creation ex. in CommandlineParsingHelper class
            See usages of set_parsed_error_handler(), set_parsed_rescue()
    """

    body: List[str]
    on_rescue: List[TaskArguments]
    on_error: List[TaskArguments]
    retry_per_task: int = 0

    _tasks: List[TaskArguments]
    _raw_attributes: dict
    _retry_counter_per_task: Dict['TaskDeclaration', int]
    _retry_counter_on_whole_block: int

    def __init__(self, body: List[str] = None, rescue: str = '', error: str = '', retry: int = 0,
                 retry_block: int = 0):
        """
        :param body Can be empty - it means that block will have tasks filled up later
        """

        if body is None:
            body = []

        self.body = body
        try:
            self.retry_per_task = int(retry)
        except ValueError:
            self.retry_per_task = 0

        try:
            self.retry_whole_block = int(retry_block)
        except ValueError:
            self.retry_whole_block = 0

        # lazy-filled by parser on later stage
        self.on_rescue = []
        self.on_error = []
        self._retry_counter_per_task = {}
        self._retry_counter_on_whole_block = 0

        # those attributes will be lazy-parsed on later processing stage
        self._raw_attributes = {
            'rescue': rescue,
            'error': error,
        }

    @staticmethod
    def from_empty() -> 'ArgumentBlock':
        """Dummy instance"""

        instance = ArgumentBlock(
            body=[], rescue='', error='', retry=0
        )

        instance.set_parsed_rescue([])
        instance.set_parsed_error_handler([])

        return instance

    def clone_with_tasks(self, tasks_arguments: List[TaskArguments]):
        cloned = deepcopy(self)
        cloned._tasks = tasks_arguments

        return cloned

    def tasks(self) -> List[TaskArguments]:
        return self._tasks

    def with_tasks_from_first_block(self, blocks: List['ArgumentBlock']):
        try:
            return self.clone_with_tasks(blocks[0]._tasks)
        except IndexError:
            return self

    def raw_attributes(self) -> dict:
        return self._raw_attributes

    def set_parsed_error_handler(self, tasks_arguments: List[TaskArguments]) -> None:
        self.on_error = tasks_arguments

    def set_parsed_rescue(self, tasks_arguments: List[TaskArguments]) -> None:
        self.on_rescue = tasks_arguments

    def should_task_be_retried(self, declaration):
        # no retry available at all
        if self.retry_per_task < 1:
            return False

        # has to retry, but it is a first time
        if declaration not in self._retry_counter_per_task:
            return True

        return self._retry_counter_per_task[declaration] < self.retry_per_task

    def task_retried(self, declaration):
        """Takes notification from external source to internally note that given task was retried"""

        if declaration not in self._retry_counter_per_task:
            self._retry_counter_per_task[declaration] = 0

        self._retry_counter_per_task[declaration] += 1

    def whole_block_retried(self, declaration):
        pass

    def should_rescue_task(self):
        """
        Decides if given task should have executed a rescue set of tasks
        """

        return len(self.on_rescue) > 0

    def has_action_on_error(self):
        """
        Answers if there is a set of tasks that should be notified on error
        """

        return len(self.on_error) > 0

    def should_block_be_retried(self) -> bool:
        """
        Can the whole block of tasks be repeated from scratch?
        """

        if self.retry_whole_block < 1:
            return False

        # actual state < declared maximum
        return self._retry_counter_on_whole_block < self.retry_whole_block

    def __str__(self):
        text = str(self.body)

        try:
            text += ', ' + str(self.tasks())
        except AttributeError:
            pass

        return 'ArgumentBlock<' + text + '>'



"""
Centralized place for environment variables supported internally by RKD
=======================================================================

Idea is to make RKD core code clean of os.getenv() calls as much as possible.
The advantage is unified list of environment variables and always the same default values.

"""

import os
from typing import List

STR_BOOLEAN_TRUE = ['true', '1', 'yes']


def rkd_paths() -> List[str]:
    return os.getenv('RKD_PATH', '').split(':')


def rkd_depth() -> int:
    return int(os.getenv('RKD_DEPTH', 0))


def binary_name() -> str:
    return os.getenv('RKD_BIN')


def distribution_name() -> str:
    return os.getenv('RKD_DIST_NAME', 'rkd')


def audit_session_log_enabled() -> bool:
    return os.getenv('RKD_AUDIT_SESSION_LOG', '').lower() in STR_BOOLEAN_TRUE


def system_log_level() -> str:
    return os.getenv('RKD_SYS_LOG_LEVEL', 'info')


def is_subprocess_compat_mode() -> bool:
    return os.getenv('RKD_COMPAT_SUBPROCESS', '').lower() in STR_BOOLEAN_TRUE



#!/usr/bin/env python3

"""
Bootstrap
=========

Initializes the application, passes arguments and environment variables from operating system
Only this layer can use sys.exit() call to pass exit code to the operating system
"""

import sys
import os
from dotenv import load_dotenv
from .execution.results import ProgressObserver
from .argparsing.parser import CommandlineParsingHelper
from .context import ContextFactory, ApplicationContext
from .resolver import TaskResolver
from .validator import TaskDeclarationValidator
from .execution.executor import OneByOneTaskExecutor
from .exception import TaskNotFoundException, ParsingException, YamlParsingException, CommandlineParsingError
from .api.inputoutput import SystemIO
from .api.inputoutput import UnbufferedStdout
from .aliasgroups import parse_alias_groups_from_env
from .packaging import find_resource_file
from rkd import env


class RiotKitDoApplication(object):
    _ctx: ApplicationContext
    _tasks_to_execute = []

    @staticmethod
    def load_environment():
        paths = env.rkd_paths()

        for path in paths:
            if os.path.isfile(path + '/.env'):
                load_dotenv(path + '/.env')

        load_dotenv(dotenv_path=os.getcwd() + '/.env')

    @staticmethod
    def make_stdout_unbuffered():
        sys.stdout = UnbufferedStdout(sys.stdout)

    @staticmethod
    def prepend_development_paths():
        """Add ./src at the beginning of PYTHONPATH - very useful for development"""

        sys.path = [os.getcwd() + '/src'] + sys.path

    def main(self, argv: list):
        if not CommandlineParsingHelper.has_any_task(argv) and not CommandlineParsingHelper.was_help_used(argv):
            self.print_banner_and_exit()

        # system wide IO instance with defaults, the :init task should override those settings
        io = SystemIO()
        io.silent = env.system_log_level() not in ['debug', 'internal']
        io.set_log_level(env.system_log_level())

        cmdline_parser = CommandlineParsingHelper(io)

        # preparse arguments that are before tasks
        preparsed_args = CommandlineParsingHelper.preparse_args(argv)

        # load context of components - all tasks, plugins etc.
        try:
            self._ctx = ContextFactory(io).create_unified_context(additional_imports=preparsed_args['imports'])

        except ParsingException as e:
            io.silent = False
            io.error_msg('Cannot import tasks/module from RKD_IMPORTS environment variable or --imports switch. '
                         'Details: {}'.format(str(e)))
            sys.exit(1)

        except YamlParsingException as e:
            io.silent = False
            io.error_msg('Cannot import tasks/module from one of makefile.yaml files. Details: {}'.format(str(e)))
            sys.exit(1)

        observer = ProgressObserver(io)
        task_resolver = TaskResolver(self._ctx, parse_alias_groups_from_env(os.getenv('RKD_ALIAS_GROUPS', '')))
        executor = OneByOneTaskExecutor(self._ctx, observer)

        # iterate over each task, parse commandline arguments
        try:
            requested_tasks = cmdline_parser.create_grouped_arguments([':init'] + argv[1:])
        except CommandlineParsingError as err:
            io.error_msg(str(err))
            sys.exit(1)

        # validate all tasks
        task_resolver.resolve(requested_tasks, TaskDeclarationValidator.assert_declaration_is_valid)

        # execute all tasks
        task_resolver.resolve(requested_tasks, executor.execute)

        executor.get_observer().execution_finished()

        sys.exit(1 if executor.get_observer().is_at_least_one_task_failing() else 0)

    @staticmethod
    def print_banner_and_exit():
        with open(find_resource_file('banner.txt'), 'rb') as banner_file:
            print(banner_file.read().replace(b'\\x1B', b'\x1B').decode('utf-8'))

        sys.exit(0)


def main():
    app = RiotKitDoApplication()
    app.make_stdout_unbuffered()
    app.prepend_development_paths()
    app.load_environment()

    try:
        app.main(argv=sys.argv)

    except TaskNotFoundException as e:
        print(e)
        sys.exit(127)


if __name__ == '__main__':
    main()



# list of tasks there is kept for compatibility, as previously "core" tasks were placed in __init__
from .core import InitTask, \
    TasksListingTask, VersionTask, \
    ShellCommandTask, LineInFileTask, \
    CallableTask, CreateStructureTask, imports as core_imports

from .env import imports as env_imports
from .jinja import imports as jinja_imports
from .shell import imports as shell_imports


def imports() -> list:
    return [] + core_imports() + jinja_imports() + shell_imports()



"""
Packaging
=========

Utilities related to finding resources of installed RKD
"""

import os
import sys
from distutils.sysconfig import get_python_lib
from typing import List, Optional, Callable
from . import env


def find_resource_directory(path: str) -> Optional[str]:
    return _find(path, os.path.isdir)


def find_resource_file(path: str) -> Optional[str]:
    return _find(path, os.path.isfile)


def get_possible_paths(path: str) -> List[str]:
    """
    Finds possible paths to resources, considering PACKAGE and USER directories first, then system-wide directories

    :param path:
    :return:
    """

    # <sphinx_resources-get_possible_paths>
    dist_name = env.distribution_name()  # RKD_DIST_NAME env variable

    paths = [
        # eg. ~/.local/share/rkd/banner.txt
        os.path.expanduser(('~/.local/share/%s/' + path) % dist_name),

        # eg. /home/andrew/.local/lib/python3.8/site-packages/rkd/misc/banner.txt
        (get_user_site_packages() + '/%s/misc/' + path) % dist_name,

        # eg. /usr/lib/python3.8/site-packages/rkd/misc/banner.txt
        (_get_global_site_packages() + '/%s/misc/' + path) % dist_name,

        # eg. /usr/share/rkd/banner.txt
        ('/usr/share/%s/' + path) % dist_name
    ]
    # </sphinx_resources-get_possible_paths>

    # eg. ./rkd/misc/banner.txt
    global_module_path = _get_current_script_path() + '/misc/' + path

    # installed module directory should be less important to allow customizations
    if "site-packages" in global_module_path:
        paths.append(global_module_path)
    else:  # local development directory
        paths = [global_module_path] + paths

    return paths


def get_user_site_packages() -> str:
    """
    Finds a user or venv site-packages directory
    :return:
    """

    try:
        return next(p for p in sys.path if 'site-packages' in p)
    except StopIteration:
        return _get_global_site_packages()


def _find(path: str, method: Callable) -> Optional[str]:
    for checked_path in get_possible_paths(path):
        if method(checked_path):
            return checked_path

    return None


def _get_global_site_packages() -> str:
    return get_python_lib()


def _get_current_script_path() -> str:
    return os.path.dirname(os.path.realpath(__file__))



#!/usr/bin/env python3

from setuptools import setup, find_packages

setup(
    setup_requires=['pbr'],
    pbr=True,
    package_dir={'': './'},
    packages=find_packages(where='./'),
    include_package_data=True,
    package_data={
        'rkd': [
            'misc/*',
            'misc/**/*',
            'misc/initial-structure/.rkd/*',
            'misc/initial-structure/.rkd/**/*',
            'misc/initial-structure/.rkd/logs/.gitkeep',
            'misc/internal/**/*'
        ]
    }
)



import importlib
from types import FunctionType
from typing import List
from ..exception import ParsingException
from .syntax import TaskDeclaration


class SyntaxParsing(object):
    @staticmethod
    def parse_imports_by_list_of_classes(classes_or_modules: List[str]) -> List[TaskDeclaration]:
        """
        Parses a List[str] of imports, like in YAML syntax.
        Produces a List[TaskDeclaration] with imported list of tasks.

        Could be used to import & validate RKD tasks.

        Examples:
            - rkd.standardlib
            - rkd.standardlib.jinja.FileRendererTask

        :raises ParsingException
        :return:
        """

        parsed: List[TaskDeclaration] = []

        for import_str in classes_or_modules:
            parts = import_str.split('.')
            class_name = parts[-1]
            import_path = '.'.join(parts[:-1])

            # importing just a full module name eg. "rkd_python"
            if len(parts) == 1:
                import_path = import_str
                class_name = 'imports'
            # Test if it's not a class name
            # In this case we treat is as a module and import an importing method imports()
            elif class_name.lower() == class_name:
                import_path += '.' + class_name
                class_name = 'imports'

            try:
                module = importlib.import_module(import_path)
            except ImportError as e:
                raise ParsingException.from_import_error(import_str, class_name, e)

            if class_name not in dir(module):
                raise ParsingException.from_class_not_found_in_module_error(import_str, class_name, import_path)

            if isinstance(module.__getattribute__(class_name), FunctionType):
                parsed += module.__getattribute__(class_name)()
            else:
                parsed.append(TaskDeclaration(module.__getattribute__(class_name)()))

        return parsed



#!/usr/bin/env python3

import os
from typing import List
from typing import Tuple
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from shlex import split as split_argv
from ..api.inputoutput import IO
from ..api.contract import TaskDeclarationInterface
from ..api.contract import ArgumentEnv
from .blocks import parse_blocks, TOKEN_BLOCK_REFERENCE_OPENING, TOKEN_BLOCK_REFERENCE_CLOSING, ArgumentBlock
from .model import TaskArguments


class TraceableArgumentParser(ArgumentParser):
    """Traces options inserted into ArgumentParser for later interpretation

    Example case: ArgumentParser does not allow to check what is the default defined value for argument
                  but we need this information somewhere, so we track add_argument() parameters
                  (it is safe as the API is stable)
    """

    traced_arguments: dict

    def __init__(self, *args, **kwargs):
        self.traced_arguments = {}
        super().__init__(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        self._trace_argument(args, kwargs)
        super().add_argument(*args, **kwargs)

    def _trace_argument(self, args: tuple, kwargs: dict):
        for arg in args:
            self.traced_arguments[arg] = {
                'default': kwargs['default'] if 'default' in kwargs else None
            }

    def get_traced_argument(self, name: str):
        return self.traced_arguments[name]


class CommandlineParsingHelper(object):
    """
    Extends argparse functionality by grouping arguments into Blocks -> Tasks arguments
    """

    io: IO

    def __init__(self, io: IO):
        self.io = io

    def create_grouped_arguments(self, commandline: List[str]) -> List[ArgumentBlock]:
        commandline, blocks = parse_blocks(commandline)

        current_group_elements = []
        current_task_name = 'rkd:initialize'
        cursor = -1
        max_cursor = len(commandline)

        parsed_into_blocks = []

        for part in commandline:
            cursor += 1

            # normalize - strip out spaces to be able to detect "-", "--" and ":" at the beginning of string
            part = part.strip()

            is_flag = part[0:1] == "-"
            is_task = part[0:1] in (':', '@')
            is_block = part.startswith(TOKEN_BLOCK_REFERENCE_OPENING) and part.endswith(TOKEN_BLOCK_REFERENCE_CLOSING)
            previous_is_flag = commandline[cursor-1][0:1] == "-" if cursor >= 1 else False

            # option name or flag
            if is_flag:
                current_group_elements.append(part)

            elif is_block:
                if part not in blocks:
                    raise Exception('Parser error. Cannot find block "{}"'.format(part))

                block: ArgumentBlock = blocks[part]
                block = block.with_tasks_from_first_block(
                    self.create_grouped_arguments(block.body)
                )

                parsed_into_blocks.append(block)

            # option value
            elif not is_flag and previous_is_flag and not is_task:
                current_group_elements.append(part)

            # new task
            elif is_task:
                if current_task_name != 'rkd:initialize':
                    task_arguments = [TaskArguments(current_task_name, current_group_elements)]

                    self.io.internal('Creating task with arguments {}'.format(task_arguments))

                    # by default every task belongs to a block, even if the block for it was not defined
                    parsed_into_blocks.append(ArgumentBlock([current_task_name] + current_group_elements)
                                              .clone_with_tasks(task_arguments))

                current_task_name = part
                current_group_elements = []

            # is not an option (--some or -s) but a positional argument actually
            else:
                current_group_elements.append(part)

            if cursor + 1 == max_cursor:
                task_arguments = [TaskArguments(current_task_name, current_group_elements)]

                self.io.internal('End of commandline arguments, closing current task collection with {}'
                                 .format(task_arguments))

                parsed_into_blocks.append(ArgumentBlock([current_task_name] + current_group_elements)
                                          .clone_with_tasks(task_arguments))

        return self._parse_shared_arguments(self.parse_modifiers_in_blocks(parsed_into_blocks))

    def parse_modifiers_in_blocks(self, blocks: List[ArgumentBlock]) -> List[ArgumentBlock]:
        """Parse list of tasks in blocks attributes eg.
        @error :notify -m 'Failed' and resolve as Notify task with -m argument"""

        for block in blocks:
            attributes = block.raw_attributes()

            if attributes['error']:
                block.set_parsed_error_handler(
                    self.create_grouped_arguments(split_argv(attributes['error']))[0].tasks())

            if attributes['rescue']:
                block.set_parsed_rescue(self.create_grouped_arguments(split_argv(attributes['rescue']))[0].tasks())

        return blocks

    @classmethod
    def _parse_shared_arguments(cls, blocks: List[ArgumentBlock]) -> List[ArgumentBlock]:
        """Apply arguments from task "@" that is before a group of tasks
           "@" without any arguments is clearing previous "@" with arguments
        """

        global_group_elements = []
        new_blocks = []

        for block in blocks:
            block_tasks = []
            block_body = []

            for task in block.tasks():
                task: TaskArguments

                if task.name() == '@':
                    global_group_elements = task.args()
                    # jump to next task - "@" task should not be finally on the list
                    continue

                block_body.append([task.name()] + task.args() + global_group_elements)
                block_tasks.append(task.with_args(task.args() + global_group_elements))

            # cut off empty blocks (ex. ['@', '--type', 'human-rights'] -> [] -> to be removed after propagation)
            if not block_body:
                continue

            # replace all blocks with new blocks that contains the additional arguments
            new_blocks.append(block.clone_with_tasks(block_tasks))

        return new_blocks

    @classmethod
    def parse(cls, declaration: TaskDeclarationInterface, args: list) -> Tuple[dict, dict]:
        """Parses ArgumentParser arguments defined by tasks

        Behavior:
          - Adds RKD-specific arguments
          - Includes task's specific arguments
          - Formats description, including documentation of environment variables

        Returns:
          Tuple of two dicts. First dict: arguments key=>value, Second dict: arguments definitions for advanced usae
        """

        argparse = TraceableArgumentParser(declaration.to_full_name(), formatter_class=RawTextHelpFormatter)

        argparse.add_argument('--log-to-file', '-rf', help='Capture stdout and stderr to file')
        argparse.add_argument('--log-level', '-rl', help='Log level: debug,info,warning,error,fatal')
        argparse.add_argument('--keep-going', '-rk', help='Allow going to next task, even if this one fails',
                              action='store_true')
        argparse.add_argument('--silent', '-rs', help='Do not print logs, just task output', action='store_true')
        argparse.add_argument('--become', '-rb', help='Execute task as given user (requires sudo)', default='')
        argparse.add_argument('--task-workdir', '-rw', help='Set a working directory for this task', default='')

        declaration.get_task_to_execute().configure_argparse(argparse)
        cls.add_env_variables_to_argparse_description(argparse, declaration)

        return vars(argparse.parse_args(args)), argparse.traced_arguments

    @classmethod
    def add_env_variables_to_argparse_description(cls, argparse: ArgumentParser, task: TaskDeclarationInterface):
        if argparse.description is None:
            argparse.description = ""

        argparse.description += task.get_full_description() + "\n"

        # print all environment variables possible to use
        argparse.description += "\nEnvironment variables for task \"%s\":\n" % task.to_full_name()

        for env in task.get_task_to_execute().internal_normalized_get_declared_envs().values():
            env: ArgumentEnv

            argparse.description += " - %s (default: %s)\n" % (
                str(env.name), str(env.default)
            )

        if not task.get_task_to_execute().get_declared_envs():
            argparse.description += ' -- No environment variables declared -- '

    @staticmethod
    def preparse_args(args: List[str]):
        """
        Parses commandline arguments which are not accessible on tasks level, but are accessible behind tasks
        Those arguments should decide about RKD core behavior on very early stage

        :param args:
        :return:
        """

        limited_args = []

        for arg in args:
            # parse everything before any task or block starts
            if arg.startswith(':') or arg.startswith('@') or arg.startswith('{@'):
                break

            limited_args.append(arg)

        argparse = ArgumentParser(add_help=False)
        argparse.add_argument('--imports', '-ri')

        parsed = vars(argparse.parse_known_args(args=limited_args)[0])

        return {
            'imports': list(filter(None,
                                   os.getenv('RKD_IMPORTS', parsed['imports'] if parsed['imports'] else '').split(':')
                                   ))
        }

    @staticmethod
    def has_any_task(argv: List[str]) -> bool:
        """
        Checks if arguments contains at least one task

        :param argv:
        :return:
        """

        for arg in argv:
            if arg.startswith(':'):
                return True

        return False

    @staticmethod
    def was_help_used(argv: List[str]) -> bool:
        for arg in argv:
            if arg in ['-h', '--help']:
                return True

        return False




import os
from pwd import getpwnam
from pickle import dumps as pickle_dumps
from pickle import loads as pickle_loads
from typing import Union, Optional
from ..argparsing.parser import CommandlineParsingHelper
from ..api.syntax import TaskDeclaration, GroupDeclaration
from ..api.contract import TaskInterface
from ..api.contract import ExecutorInterface
from ..api.contract import ExecutionContext
from ..context import ApplicationContext
from ..inputoutput import IO
from ..inputoutput import SystemIO
from ..inputoutput import output_formatted_exception
from ..process import switched_workdir
from ..exception import InterruptExecution, \
    ExecutionRetryException, \
    ExecutionRescueException, \
    ExecutionErrorActionException
from .results import ProgressObserver
from ..audit import decide_about_target_log_files
from ..api.temp import TempManager
from .serialization import FORKED_EXECUTOR_TEMPLATE
from .serialization import get_unpicklable


class OneByOneTaskExecutor(ExecutorInterface):
    """ Executes tasks one-by-one, providing a context that includes eg. parsed arguments """

    _ctx: ApplicationContext
    _observer: ProgressObserver
    io: SystemIO

    def __init__(self, ctx: ApplicationContext, observer: ProgressObserver):
        self._ctx = ctx
        self.io = ctx.io
        self._observer = observer

    def execute(self, declaration: TaskDeclaration, task_num: int, parent: Union[GroupDeclaration, None] = None,
                args: list = None):

        """
        Executes a single task passing the arguments, redirecting/capturing the output and handling the errors
        """

        if args is None:
            args = []

        # 1. notify
        self._observer.task_started(declaration, parent, args)

        # 2. parse arguments
        parsed_args, defined_args = CommandlineParsingHelper.parse(declaration, args)
        log_level: str = parsed_args['log_level']
        log_to_file: str = parsed_args['log_to_file']
        is_silent: bool = parsed_args['silent']
        keep_going: bool = parsed_args['keep_going']
        cmdline_become: str = parsed_args['become']
        workdir = parsed_args.get('task_workdir') if parsed_args.get('task_workdir') else declaration.workdir

        # 3. execute
        temp = TempManager()

        try:
            io = IO()
            io.set_log_level(log_level if log_level else self.io.get_log_level())

            if is_silent:
                io.silent = is_silent
            else:
                io.inherit_silent(self.io)  # fallback to system-wide

            where_to_store_logs = decide_about_target_log_files(self._ctx, log_to_file, declaration, task_num)

            with io.capture_descriptors(target_files=where_to_store_logs):

                task = declaration.get_task_to_execute()
                task.internal_inject_dependencies(io, self._ctx, self, temp)

                with switched_workdir(workdir):
                    result = self._execute_directly_or_forked(cmdline_become, task, temp, ExecutionContext(
                            declaration=declaration,
                            parent=parent,
                            args=parsed_args,
                            env=declaration.get_env(),
                            defined_args=defined_args
                        ))

        # 4. capture result
        except Exception as e:
            #
            # When: Task has a failure
            #
            temp.finally_clean_up()
            self._on_failure(declaration, keep_going, e, parent)

            return

        #
        # When: Task did not raise exception
        #
        temp.finally_clean_up()

        if result is True:
            self._observer.task_succeed(declaration, parent)
        else:
            self._on_failure(declaration, keep_going, None, parent)

    def _on_failure(self, declaration: TaskDeclaration, keep_going: bool,
                    exception: Optional[Exception] = None,
                    parent: Union[GroupDeclaration, None] = None):

        """
        Executed when task fails - regardless of if it is an Exception raised or just a False returned

        Roles:
            - Block: a domain logic, tracks TaskDeclaration execution
              (only those declarations that are declared in that block)
            - Observer: Observes execution RESULTS to notify user, the console (to set exit code for example).
                        Needs also to be notified, when tasks are retried and when those retried tasks are passing
                        eg. second time ater failing first time
        """

        # block modifiers (issue #50): @retry a task up to X times
        if declaration.block().should_task_be_retried(declaration):
            declaration.block().task_retried(declaration)
            self._observer.task_retried(declaration)

            raise ExecutionRetryException() from exception

        elif declaration.block().should_block_be_retried():
            declaration.block().whole_block_retried(declaration)

            self._observer.group_of_tasks_retried(declaration.block())

            raise ExecutionRetryException(declaration.block().tasks()) from exception

        # regardless of @error & @rescue there should be a visible information that task failed
        self._notify_error(declaration, parent, exception)

        # block modifiers (issue #50): @error and @rescue
        if declaration.block().should_rescue_task():
            self._observer.task_rescue_attempt(declaration)

            raise ExecutionRescueException(declaration.block().on_rescue) from exception

        elif declaration.block().has_action_on_error():
            raise ExecutionErrorActionException(declaration.block().on_error) from exception

        # break the whole pipeline only if not --keep-going
        if not keep_going:
            raise InterruptExecution() from exception

    def _notify_error(self, declaration: TaskDeclaration,
                      parent: Union[GroupDeclaration, None] = None,
                      exception: Optional[Exception] = None):

        """Write to console, notify observers"""

        # distinct between error and failure, first has a stacktrace, second is a logical task failure
        if not exception:
            self._observer.task_failed(declaration, parent)
        else:
            output_formatted_exception(exception, str(declaration.get_task_to_execute().get_full_name()), self.io)
            self._observer.task_errored(declaration, exception)

    def _execute_directly_or_forked(self, cmdline_become: str, task: TaskInterface, temp: TempManager, ctx: ExecutionContext):
        """Execute directly or pass to a forked process
        """

        if task.should_fork() or cmdline_become:
            task.io().debug('Executing task as separate process')
            return self._execute_as_forked_process(cmdline_become, task, temp, ctx)

        return task.execute(ctx)

    @staticmethod
    def _execute_as_forked_process(become: str, task: TaskInterface, temp: TempManager, ctx: ExecutionContext):
        """Execute task code as a separate Python process

        The communication between processes is with serialized data and text files.
        One text file is a script, the task code is passed with stdin together with a whole context
        Second text file is a return from executed task - it can be a boolean or exeception.

        When an exception is returned by a task, then it is reraised there - so the original exception is shown
        without any proxies.
        """

        if not become:
            become = task.get_become_as()

        # prepare file with source code and context
        communication_file = temp.assign_temporary_file()
        task.io().debug('Assigning communication temporary file at "%s"' % communication_file)

        context_to_pickle = {'task': task, 'ctx': ctx, 'communication_file': communication_file}

        try:
            task.io().debug('Serializing context')
            with open(communication_file, 'wb') as f:
                f.write(pickle_dumps(context_to_pickle))

        except (AttributeError, TypeError) as e:
            task.io().error('Cannot fork, serialization failed. ' +
                            'Hint: Tasks that are using internally inner-methods and ' +
                            'lambdas cannot be used with become/fork')
            task.io().error(str(e))

            if task.io().is_log_level_at_least('debug'):
                task.io().error('Pickle trace: ' + str(get_unpicklable(context_to_pickle)))

            return False

        # set permissions to temporary file
        if become:
            task.io().debug('Setting temporary file permissions')
            os.chmod(communication_file, 0o777)

            try:
                getpwnam(become)
            except KeyError:
                task.io().error('Unknown user "%s"' % become)
                return False

        task.io().debug('Executing python code')
        task.py(code=FORKED_EXECUTOR_TEMPLATE, become=become, capture=False, arguments=communication_file)

        # collect, process and pass result
        task.io().debug('Parsing subprocess results from a serialized data')
        with open(communication_file, 'rb') as conn_file:
            task_return = pickle_loads(conn_file.read())

        if isinstance(task_return, Exception):
            task.io().debug('Exception was raised in subprocess, re-raising')
            raise task_return

        return task_return

    def get_observer(self) -> ProgressObserver:
        return self._observer
